#! /usr/bin/python3

__doc__ = """\
RPM Build Python3 / Compileall 
Converted from RPM Build Python  / Compile All by 2to3 script

Description:

    This application is intendent to byte compile python3 modules 

Call:
    compileall.py <prefix>

Switches:

    --exclude=<LIST> -- List of paths to be excluded from byte compilation;

    --include=<LIST> -- List of paths to be included into byte compilation;

    --deep=<INT> -- Deep of search down into directory tree;

    --skip_x -- Skip executable files from byte compilation;

    --clean -- Clean not-comileable modules;

    --prefix=<STRING> -- Install prefix;

Warranty:

    There is no warranty of any kind, either expressed
    or implied

TODO:
    Nothing

"""
def _(s) : return s

import sys, os, py_compile

ALL = 1
ALREADY = 2
BOTH = 3
SUFFIX = __debug__ and 'c' or 'o'

def join_prefix(prefix,path) :
    return os.path.abspath(os.path.join(prefix,os.path.abspath(path)[1:]))

def executable(file):
    import stat
    if stat.S_IEXEC & (os.stat(file)[stat.ST_MODE]) :
        header = open(file, 'rb').read(2)
        if b'#!' == header :
            return True

def compile_all(path,exclude=['/usr/share/doc'],deep=20,skip_x=False,clean=False,prefix='') :
    if path not in exclude and deep > 0 :
        try:
            lp = os.listdir(path)
        except os.error:
            print(_("Can't list '%s'") % path)
        else :
            lp.sort()
            lp = [ os.path.join(path,p) for p in lp ]

            for p in lp :
                if os.path.isdir(p) :
                    compile_all(p,exclude,deep-1,skip_x,clean,prefix)

            if clean :
                for p in lp :
                    if os.path.isfile(p) and os.path.splitext(p)[1] == ('.py' + SUFFIX) :
                        print(_('unlink'),p)
                        os.unlink(p)

            for p in lp :
                if os.path.isfile(p) :
                    if os.path.splitext(p)[1] == '.py' :
                        if not (skip_x and executable(p)) :
                            try :
                                print(_('compile'),p)
                                py_compile.compile(p, doraise=True)
                            except py_compile.PyCompileError as msg:
                                print(_("%s can't be compiled : %s") % (p,msg))

def compile_path(include=sys.path[2:],exclude=[],deep=10,skip_x=False,clean=False,prefix='/') : 
    for item in set(include) :
        compile_all(join_prefix(prefix,item),[join_prefix(prefix,x) for x in exclude],deep,skip_x,clean,prefix)

def main() :
    """Script main program."""
    import getopt,re
    lopt = ['exclude=','include=','deep=','skip_x','clean','prefix=']
    cut = re.compile("^-*([^=]*)=?$").match
    lenv = [ ('python_bc_%s' % cut(x).group(1)) for x in lopt ]
    dopt = dict([ (x[0:10],y) for x,y in list(os.environ.items()) if x in lenv])

    try:
        opts, args = getopt.getopt(sys.argv[1:], 'h',lopt)
    except getopt.error as msg:
        print(msg)
        print(__doc__)
        sys.exit(2)
    else :
        dopt.update(dict([ (cut(x).group(1),y) for x,y in opts]))
        if 'h' in dopt :
            print(__doc__)
        else :
            if 'exclude' in dopt :
                dopt['exclude'] = dopt['exclude'].split()
            if 'include' in dopt and dopt['include'] :
                dopt['include'] = dopt['include'].split()
            if 'skip_x' in dopt :
                dopt['skip_x'] = True
            if 'clean' in dopt :
                dopt['clean'] = True
            if 'deep' in dopt :
                dopt['deep'] = int(dopt['deep'])

            if args :
                dopt['include'] = args

            return compile_path(**dopt)

if __name__ == '__main__' :
    exit_status = bool(main())
    sys.exit(exit_status)
