#! /usr/bin/python3

import ast
import sys
import os
import re

def is_import_stmt(node):
    return type(node) == ast.Call and type(node.func) == ast.Name and node.func.id == '__import__'

def is_importlib_call(node):
    return type(node) == ast.Call and type(node.func) == ast.Attribute and type(node.func.value) == ast.Name and node.func.value.id == 'importlib'

def requ_check(code, path): #For REQ in ['slight', 'relaxed']
    dep = set()
    relative_dep = set()
    for warm in ast.parse(code).body:
        if type(warm) == ast.Import:
            for j in warm.names:
                dep.add(j.name)
        elif type(warm) == ast.ImportFrom:
            if warm.level != 0:
                if behavior != 'original':
                    dep.add(parents_package(path, warm.level))
            else:
                dep.add(warm.module)
        elif 'value' in dir(warm) and is_import_stmt(warm.value):
            try:
                if type(warm.value.args[0]) == ast.BinOp and type(warm.value.args[0].left.value) != ast.Attribute:
                    relative_dep.add(warm.value.args[0].left.value)
                elif type(warm.value.args[0].value) != ast.Name and type(warm.args[0].value) == str:
                    relative_dep.add(warm.value.args[0].value)
            except Exception as msg:
                print(msg, file=sys.stderr)
                pass
    return dep, relative_dep

def parents_package(path, level):
    parent = ''
    path = (os.path.abspath(path))[:-len(os.path.basename(path)) - 1]

    for i in range(level):
        head, tail = os.path.split(path)
        parent = tail + '.' + parent
        path = head
    return parent[:-1]

def read_ast_tree(codes, path, REQ):
    dep = set()
    relative_dep = set()
    try:
        filter_count = 0
        for chuck_norris in ast.walk(ast.parse(codes)):
            if type(chuck_norris) == ast.Import:
                for j in chuck_norris.names:
                    dep.add(j.name)
            elif type(chuck_norris) == ast.ImportFrom:
                if sys.argv.count('--include-self-provides') == 0 and chuck_norris.level != 0:
                    print('python3.req.py:%s: possible relative import from %s' %(path, parents_package(path, chuck_norris.level)), file=sys.stderr)
                elif chuck_norris.level != 0:
                    dep.add(parents_package(path, chuck_norris.level))
                else:
                    dep.add(chuck_norris.module)
            elif is_import_stmt(chuck_norris):
                try:
                    if type(chuck_norris.args[0]) == ast.BinOp and type(chuck_norris.args[0].left.value) != ast.Attribute:
                        relative_dep.add(chuck_norris.args[0].left.value)
                    elif type(chuck_norris.args[0].value) != ast.Name and type(chuck_norris.args[0].value) == str:
                        relative_dep.add(chuck_norris.args[0].value)
                    else:
                        raise
                except Exception as msg:
                    print(msg, file=sys.stderr)
                    if sys.argv.count('--verbose') != 0:
                        print('python3.req.py:%s: failed to determine the name of the module contained in the variable: %s' % (path, chuck_norris.args[0].id), file=sys.stderr)
                        print('python3.req.py:%s: a form of __import__ was read which gave no deps\n%s' % (path, ast.dump(chuck_norris, indent=4)), file=sys.stderr)
            elif sys.argv.count('--importlib') != 0 and is_importlib_call(chuck_norris):
                dep.add(chuck_norris.args[0].value)
    except SyntaxError:
        print('python3.req: error: %s: invalid syntax' % (path), file=sys.stderr)
        if file_name[-3:] == '.py' or len(code) > 0 and code[0].startswith('#!'):
            pass
        else:
            print('python3.req: %s: possibly not pythonish file' % file_name, file=sys.stderr)
    return dep, relative_dep

#TODO: need to detect provides, when use python3.req.py IRL just for 1 file
#def scanner(path): #Maybe for future

def provides_creator(file_path):
    provides = []

    if sys.argv.count('--include-self-provides') != 0 or behavior == 'experiemental':
        provides = scanner(file_path)
    else:
        file_path, trash = os.path.splitext(file_path)
        file_path = file_path.split('/')
        file_path.reverse()
        if '' in file_path:
            file_path.remove('')
        for material in file_path:
            provides.append(material + '.' + provides[-1] if len(provides) > 0 else material)
        if file_path[0] == '__init__':
            provides_material = provides.copy()
            for material in provides_material:
                material = '.'.join(material.split('.')[:-1])
                if len(material) > 0:
                    provides.append(material)
    return provides

def catch_so(name):
    dep_version = os.getenv('RPM_PYTHON3_VERSION', "%s.%s" % sys.version_info[0:2])

    import subprocess
    res = subprocess.run(['objdump', '-p', name], capture_output=True, stderr=None)
    if res.returncode != 0:
        print(res.stderr.decode(), file=sys.stderr)
        return
    if res.stdout.find(b'file format elf64') != -1: 
        dep_suffix = "(64bit)"
    else:
        dep_suffix = ""
    print("python%s-ABI%s" % (dep_version, dep_suffix))

def ignore(name):
    if IS_HIER == None:
        name = name.split('.')[0]
    if name in dict([(x, 1) for x in list(os.getenv('RPM_PYTHON3_REQ_SKIP', "").split()) + list(sys.builtin_module_names)]):
        return True
    else:
        return False

def get_raw_encoding(line) :
    enc = re_coding.search(line)
    if enc :
        return enc.group(1)

# Parser/tokenizer.c:get_normal_name()
def get_enc_normal_name(enc) :
    enc = enc.lower().replace('_','-')
    if enc.startswith('utf-8') :
        return 'utf-8'
    for alias in ['latin-1','iso-8859-1','iso-latin-1'] :
        if enc.startswith(alias) :
            return 'iso-8859-1'
    return enc

missing_encodings = None
def get_encoding_module(enc, file_name):
    mod = 'encodings'
    try :
        from encodings import search_function
    except Exception as msg:
        print(msg, file=sys.stderr)
        global missing_encodings
        missing_encodings = True
        return mod
    try :
        mod = search_function(enc)[1].__self__.__class__.__module__
        print("python3.req: %s: encoding=%s module=%s" % (file_name,enc,mod), file=sys.stderr)
    except Exception as msg:
        print(msg, file=sys.stderr)
        pass
    return mod

def need_encoding_module(line, file_name):
    enc = get_raw_encoding(line)
    if enc :
        enc = get_enc_normal_name(enc)
        if enc not in ['utf-8','iso-8859-1'] :
            print("python3.req: %s: non-standard encoding: %s" % (file_name,enc), file=sys.stderr)
            mod = get_encoding_module(enc, file_name)
            return mod

def printing(dependencies, filters , relative_flag, file_name, REQ):
    head, tail = os.path.split(file_name)
    for dep in dependencies:
        if ignore(dep): #Исключаются те модули, которые либо уже предоставлены конкретной данной сборкой python3, либо исключены средой сборки 
            print("python3.req: %s: skipping %s" % (file_name, dep), file=sys.stderr)
        elif not (dep in filters) and REQ in ['slight', 'relaxed']:
            print('%s: %s: Ignore for REQ=%s module=%s' % (sys.argv[0], file_name, REQ, dep), file=sys.stderr)
        elif relative_flag == False:#Если зависимости были получены не __import__
            if IS_HIER == None:
                dep = str(dep).split('.')[0]
            if str(dep) in provides: 
                continue
            print("%s(%s)" % ("python%u" % sys.version_info[0], dep), end ='\n')
        elif relative_flag:#Если зависимсти были получены __import__
            if str(dep) in provides: 
                continue
            else:
                print("%s(%s)" % ("python%u" % sys.version_info[0], dep), end ='\n')
        else:
            print('UNEXPECTED SITUATION\nERROR 437', end ='\n', file = sys.stderr)

behavior = os.getenv('RPM_PYTHON3_BEHAVIORS', '')
if behavior == '':
    behavior = 'original'

REQ = os.getenv('RPM_PYTHON3_REQ_METHOD','strict')
IS_HIER = os.getenv('RPM_PYTHON3_REQ_HIER',None) #Штучка определяет для Import и FromImport завимости с или без сабмодулей
if IS_HIER and IS_HIER.startswith('no') or sys.argv.count('--no-submodules') != 0:
    # rpm-build-python/tools/requires.py uses 'hier' as the default value;
    # rpm-build-python3/python3 can set it to 'yes'.
    # So, let's suppose the opposite value can be either 'nohier' or 'no'
    # (or simply an unset env variable). Here we map those possibilities
    # to our internal canonical representation of this value
    # (just in case someone may use them).
    IS_HIER = None

file_list = []
if len(sys.argv) > 1:
    for x in sys.argv[1:]:
        if x[:2] != '--':
            file_list.append(x)
else:
    file_list = sys.stdin.read().split()

re_coding = re.compile(r'^\s*#.*?coding[:=]\s*([\w._-]+)')
provides = []
for file_name in file_list:
    provides += provides_creator(file_name)

for file_name in file_list:
    if str(file_name)[-3:] == '.so':
        catch_so(file_name)
    elif not str(file_name)[-4:] in ['.pyc', '.pyo', '.pth']:
        try:
            f = open(file_name, encoding="utf-8-sig")
        except PermissionError:
            continue
        except FileNotFoundError:
            continue
        except IsADirectoryError:
            continue

        try:
            lines = []
            for x in f:
                lines.append(x)
        except Exception as msg:
            print(msg, file=sys.stderr)
            lines = []
        f.close()
 
        lines.append('')
        code = ''.join(lines)
        dependencies, relative_dependencies = read_ast_tree(code, file_name, REQ)
        if sys.argv.count('--no-encodings') == 0:
            encmod = None
            if len(lines) > 0 :
                encmod = need_encoding_module(lines[0], file_name)
            if not encmod and len(lines) > 1 :
                encmod = need_encoding_module(lines[1], file_name)
            if encmod:
                if sys.argv.count('--include-encodings') != 0 or behavior == 'original':
                    print("%s(%s)" % ("python%u" % sys.version_info[0], encmod), end ='\n')
                else:
                    dependencies.add(encmod)
        
        filtered_rel_dep = []
        filtered_dep = []
        if REQ in ['slight', 'relaxed']:
            filtered_dep, filtered_rel_dep = requ_check(code, file_name)
        
        if behavior == 'original':
            dependencies = sorted(dependencies)
            relative_dependencies = sorted(relative_dependencies)

        printing(dependencies, filtered_dep, False, file_name, REQ)
        printing(relative_dependencies, filtered_rel_dep, True, file_name, REQ)
